////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.send.flickr;

import java.io.UnsupportedEncodingException;
import java.util.Calendar;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.json.JSONObject;
import org.watto.android.xml.XMLNode;
import org.watto.android.xml.XMLReader;
import org.watto.program.android.send.flickr.authenticator.FlickrAuthenticator;
import org.watto.program.android.send.flickr.http.HttpMultipartMode;
import org.watto.program.android.send.flickr.http.MultipartEntity;
import org.watto.program.android.send.flickr.http.content.ByteArrayBody;
import org.watto.program.android.send.flickr.http.content.StringBody;


/***********************************************************************************************
Provides utility methods for communicating with the backend server
***********************************************************************************************/
public class FlickrQuery {

  /***********************************************************************************************
  Fetches the username of the authenticated user from the server
  @return The username
  ***********************************************************************************************/
  public static String getUsername(){
    try {

      if (!FlickrAuthenticator.credentialsKnown()) {
        // The authentication is invalid, so return nothing
        return null;
      }

      // Get the username of the logged-in user
      JSONObject userObject = queryServer("http://api.flickr.com/services/rest/?method=flickr.test.login");
      String username = userObject.getJSONObject("user").getJSONObject("username").optString("_content",null);
      return username;

    }
    catch (Throwable t) {
      // An error occurred in the response, or there were parsing errors
      return null;
    }
  }


  /***********************************************************************************************
  Sends a POST request to the server and returns the response message
  @param request the request to perform
  @return the response, or <b>null</b> if there was an error
  ***********************************************************************************************/
  public static String postServer(HttpPost request){
    try {

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
        // Success
        String responseMessage = EntityUtils.toString(response.getEntity());

        // Crop off the "jsonFlickrApi(" response header
        int rspPosition = responseMessage.indexOf("<rsp");
        if (rspPosition <= 0 || responseMessage.length() < rspPosition) {
          return null;
        }
        // return the response
        return responseMessage.substring(rspPosition);
      }
      return null;
    }
    catch (Throwable t) {
      return null;
    }
  }


  /***********************************************************************************************
  Signs a query, sends it to the server, receives a response and converts it to a <code>JSONObject</code> 
  @param queryUrl the URL of the query you wish to perform
  @return the response as a <code>JSONObject</code>, or <b>null</b> if there was an error
  ***********************************************************************************************/
  public static JSONObject queryServer(String queryUrl){
    try {

      // Check that we have valid authentication
      if (!FlickrAuthenticator.credentialsKnown()) {
        return null;
      }

      // Sign the request with the access token
      queryUrl = FlickrAuthenticator.signRequest(queryUrl);
      HttpGet request = new HttpGet(queryUrl);

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
        // Success
        String responseMessage = EntityUtils.toString(response.getEntity());

        // Crop off the "jsonFlickrApi(" response header
        int bracketPosition = responseMessage.indexOf('(') + 1;
        if (bracketPosition <= 1 || responseMessage.length() < bracketPosition) {
          return null;
        }
        responseMessage = responseMessage.substring(bracketPosition);

        // return the JSONObject for the response
        return new JSONObject(responseMessage);
      }
      return null;
    }
    catch (Throwable t) {
      return null;
    }
  }


  /***********************************************************************************************
  Uploads a photo to the server
  @param photoBytes the photo data as a <code>byte</code> array
  @param title the title of the photo
  @param description the description of the photo
  @param tags the tags that describe the photo
  @param visibility who has access to view the photo
  @param safety the safety level of the photo content
  @param contantType the type of image being uploaded
  @return <b>true</b> if the photo was uploaded successfully<br />
          <b>false</b> if there was an error or problem uploading the photo
  ***********************************************************************************************/
  public static boolean uploadPhoto(byte[] photoBytes,String title,String description,String tags,String visibility,String safety,String contentType){
    try {

      if (!FlickrAuthenticator.credentialsKnown()) {
        // The authentication is invalid, so return nothing
        return false;
      }

      // Encode all the String-based arguments in URLEncoding
      //title = URLEncoder.encode(title);
      //description = URLEncoder.encode(description);
      //tags = URLEncoder.encode(tags);

      // convert all the flag strings to numbers
      int isPublic = 0;
      int isFriend = 0;
      int isFamily = 0;
      int safetyNumber = 1;
      int contentTypeNumber = 1;
      int isHidden = 1;

      if (visibility.equals("public")) {
        isPublic = 1;
        isFriend = 1;
        isFamily = 1;
        isHidden = 1;
      }
      else if (visibility.equals("friends")) {
        isPublic = 0;
        isFriend = 1;
        isFamily = 0;
        isHidden = 2;
      }
      else if (visibility.equals("family")) {
        isPublic = 0;
        isFriend = 0;
        isFamily = 1;
        isHidden = 2;
      }
      else if (visibility.equals("friends_family")) {
        isPublic = 0;
        isFriend = 1;
        isFamily = 1;
        isHidden = 2;
      }
      else if (visibility.equals("private")) {
        isPublic = 0;
        isFriend = 0;
        isFamily = 0;
        isHidden = 2;
      }

      if (safety.equals("safe")) {
        safetyNumber = 1;
      }
      else if (safety.equals("moderate")) {
        safetyNumber = 2;
      }
      else if (safety.equals("restricted")) {
        safetyNumber = 3;
      }

      if (contentType.equals("photo")) {
        contentTypeNumber = 1;
      }
      else if (contentType.equals("screenshot")) {
        contentTypeNumber = 2;
      }
      else if (contentType.equals("other")) {
        contentTypeNumber = 3;
      }

      // Create a new MultipartEntity for holding the POST data
      MultipartEntity photoEntity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
      //ListeningMultipartEntity photoEntity = new ListeningMultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE,progressListener);

      // Add the photo data
      String filename = generatePhotoFilename();
      ByteArrayBody photoData = new ByteArrayBody(photoBytes,"image/jpeg",filename);
      photoEntity.addPart("photo",photoData);


      // Build the URL with all the arguments attached to it
      boolean firstArgument = true;
      String uploadUrl = "http://api.flickr.com/services/upload/?";
      if (title != null && !title.equals("")) {
        if (!firstArgument) {
          uploadUrl += "&";
        }
        uploadUrl += "title=" + title;
        firstArgument = false;
      }
      if (description != null && !description.equals("")) {
        if (!firstArgument) {
          uploadUrl += "&";
        }
        uploadUrl += "description=" + description;
        firstArgument = false;
      }
      if (tags != null && !tags.equals("")) {
        if (!firstArgument) {
          uploadUrl += "&";
        }
        uploadUrl += "tags=" + tags;
        firstArgument = false;
      }

      if (!firstArgument) {
        uploadUrl += "&";
      }
      uploadUrl += "is_public=" + isPublic + "&is_friend=" + isFriend + "&is_family=" + isFamily + "&safety_level=" + safetyNumber + "&content_type=" + contentTypeNumber + "&hidden=" + isHidden;


      // Sign the request with the access token, and generate the signature
      uploadUrl = FlickrAuthenticator.signRequest(uploadUrl);

      // Now that we have all the arguments and have calculated the signature, remove all
      // the arguments from the URL and add them to the POST MultipartEntity instead
      uploadUrl = convertUrlToMultipart(uploadUrl,photoEntity);


      // Set the request data
      HttpPost request = new HttpPost(uploadUrl);
      request.setEntity(photoEntity);

      // Upload the photo to the server (execute the request)
      String response = postServer(request);

      try {
        XMLNode responseNode = XMLReader.read(response);
        
        if (responseNode != null){
          // check that the response was "ok"
          String uploadSuccess = responseNode.getAttribute("stat");
          if (uploadSuccess != null && uploadSuccess.equals("ok")){
            // Successful upload
            return true;
          }

        }
        return false;
      }
      catch (Throwable t) {
        // An error occurred in the response, or there were parsing errors
        return false;
      }

    }
    catch (Throwable t) {
      // An error occurred in the response, or there were parsing errors
      return false;
    }
  }


  /***********************************************************************************************
  Generates a filename for the photo. The filename is a timestamp
  @return a filename in the format YYYY-MM-DD_HH.MM.SS.jpg
  ***********************************************************************************************/
  public static String generatePhotoFilename(){
    Calendar date = Calendar.getInstance();
    return date.get(Calendar.YEAR) + "-" + date.get(Calendar.MONTH) + "-" + date.get(Calendar.DAY_OF_MONTH) + "_" + date.get(Calendar.HOUR_OF_DAY) + "." + date.get(Calendar.MINUTE) + "." + date.get(Calendar.SECOND) + ".jpg";
  }


  /***********************************************************************************************
  Removes all the arguments from a URL and adds them as parts of a Multipart POST entity
  @param url the url with arguments to convert
  @param entity the <code>MultipartEntity</code> to store the argument parts
  @return the <i>url</i> without the arguments
  ***********************************************************************************************/
  public static String convertUrlToMultipart(String url,MultipartEntity entity){

    // Get all the arguments of the URL, and split them up around the '&' sign
    int questionMarkPosition = url.indexOf('?');
    String argumentsOnly = url.substring(questionMarkPosition + 1);
    String[] arguments = argumentsOnly.split("&");

    int argumentCount = arguments.length;
    for (int i = 0;i < argumentCount;i++) {
      String argument = arguments[i];

      // Split each argument up around the '=' sign to get the name and value
      int equalPosition = argument.indexOf('=');
      if (equalPosition > 0) {
        String name = argument.substring(0,equalPosition);
        String value = argument.substring(equalPosition + 1);
        try {
          // add the argument as a part in the multipart entity
          entity.addPart(name,new StringBody(value));
        }
        catch (UnsupportedEncodingException e) {
          // Just in case there is a problem in encoding the value (should never happen)
        }
      }
    }

    // return the URL without the arguments
    return url.substring(0,questionMarkPosition);
  }

}
